#if (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7,1))
#define COMPAT_HAVE_SMB_MB__AFTER_ATOMIC
#endif
#if (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7,3))
#define COMPAT_WAIT_FUNCTION_HAS_2_PARAM
#define COMPAT_HAVE_WAIT_ON_BIT_LOCK_ACTION
#endif
#if (RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7,4))
#define COMPAT_BIO_OPS_BACKPORTED
#endif
